################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from typing import Dict, Union, List

from pyflink.common import SerializationSchema, DeserializationSchema, \
    AssignerWithPeriodicWatermarksWrapper
from pyflink.datastream.functions import SourceFunction
from pyflink.datastream.connectors import Sink
from pyflink.java_gateway import get_gateway

__all__ = [
    'KinesisShardAssigner',
    'KinesisDeserializationSchema',
    'WatermarkTracker',
    'PartitionKeyGenerator',
    'FlinkKinesisConsumer',
    'KinesisStreamsSink',
    'KinesisStreamsSinkBuilder',
    'KinesisFirehoseSink',
    'KinesisFirehoseSinkBuilder'
]


# ---- KinesisSource ----


class KinesisShardAssigner(object):
    """
    Utility to map Kinesis shards to Flink subtask indices. Users can provide a Java
    KinesisShardAssigner in Python if they want to provide custom shared assigner.
    """
    def __init__(self, j_kinesis_shard_assigner):
        self._j_kinesis_shard_assigner = j_kinesis_shard_assigner

    @staticmethod
    def default_shard_assigner() -> 'KinesisShardAssigner':
        """
        A Default KinesisShardAssigner that maps Kinesis shard hash-key ranges to Flink subtasks.
        """
        return KinesisShardAssigner(get_gateway().jvm.org.apache.flink.streaming.connectors.
                                    kinesis.internals.KinesisDataFetcher.DEFAULT_SHARD_ASSIGNER)

    @staticmethod
    def uniform_shard_assigner() -> 'KinesisShardAssigner':
        """
        A KinesisShardAssigner that maps Kinesis shard hash-key ranges to Flink subtasks.
        It creates a more uniform distribution of shards across subtasks than org.apache.flink. \
        streaming.connectors.kinesis.internals.KinesisDataFetcher.DEFAULT_SHARD_ASSIGNER when the
        Kinesis records in the stream have hash keys that are uniformly distributed over all
        possible hash keys, which is the case if records have randomly-generated partition keys.
        (This is the same assumption made if you use the Kinesis UpdateShardCount operation with
        UNIFORM_SCALING.)
        """
        return KinesisShardAssigner(get_gateway().jvm.org.apache.flink.streaming.connectors.
                                    kinesis.util.UniformShardAssigner())


class KinesisDeserializationSchema(object):
    """
    This is a deserialization schema specific for the Flink Kinesis Consumer. Different from the
    basic DeserializationSchema, this schema offers additional Kinesis-specific information about
    the record that may be useful to the user application.
    """

    def __init__(self, j_kinesis_deserialization_schema):
        self._j_kinesis_deserialization_schema = j_kinesis_deserialization_schema


class WatermarkTracker(object):
    """
    The watermark tracker is responsible for aggregating watermarks across distributed operators.
    It can be used for sub tasks of a single Flink source as well as multiple heterogeneous sources
    or other operators.The class essentially functions like a distributed hash table that enclosing
    operators can use to adopt their processing / IO rates
    """

    def __init__(self, j_watermark_tracker):
        self._j_watermark_tracker = j_watermark_tracker

    @staticmethod
    def job_manager_watermark_tracker(
            aggregate_name: str, log_accumulator_interval_millis: int = -1) -> 'WatermarkTracker':
        j_watermark_tracker = get_gateway().jvm.org.apache.flink.streaming.connectors.kinesis.util \
            .JobManagerWatermarkTracker(aggregate_name, log_accumulator_interval_millis)
        return WatermarkTracker(j_watermark_tracker)


class FlinkKinesisConsumer(SourceFunction):
    """
    The Flink Kinesis Consumer is an exactly-once parallel streaming data source that subscribes to
    multiple AWS Kinesis streams within the same AWS service region, and can handle resharding of
    streams. Each subtask of the consumer is responsible for fetching data records from multiple
    Kinesis shards. The number of shards fetched by each subtask will change as shards are closed
    and created by Kinesis.

    To leverage Flink's checkpointing mechanics for exactly-once streaming processing guarantees,
    the Flink Kinesis consumer is implemented with the AWS Java SDK, instead of the officially
    recommended AWS Kinesis Client Library, for low-level control on the management of stream state.
    The Flink Kinesis Connector also supports setting the initial starting points of Kinesis
    streams, namely TRIM_HORIZON and LATEST.

    Kinesis and the Flink consumer support dynamic re-sharding and shard IDs, while sequential,
    cannot be assumed to be consecutive. There is no perfect generic default assignment function.
    Default shard to subtask assignment, which is based on hash code, may result in skew, with some
    subtasks having many shards assigned and others none.

    It is recommended to monitor the shard distribution and adjust assignment appropriately.
    A custom assigner implementation can be set via setShardAssigner(KinesisShardAssigner) to
    optimize the hash function or use static overrides to limit skew.

    In order for the consumer to emit watermarks, a timestamp assigner needs to be set via
    setPeriodicWatermarkAssigner(AssignerWithPeriodicWatermarks) and the auto watermark emit
    interval configured via ExecutionConfig.setAutoWatermarkInterval(long).

    Watermarks can only advance when all shards of a subtask continuously deliver records.
    To avoid an inactive or closed shard to block the watermark progress, the idle timeout should
    be configured via configuration property ConsumerConfigConstants.SHARD_IDLE_INTERVAL_MILLIS.
    By default, shards won't be considered idle and watermark calculation will wait for newer
    records to arrive from all shards.

    Note that re-sharding of the Kinesis stream while an application (that relies on the Kinesis
    records for watermarking) is running can lead to incorrect late events. This depends on how
    shards are assigned to subtasks and applies regardless of whether watermarks are generated in
    the source or a downstream operator.
    """

    def __init__(self,
                 streams: Union[str, List[str]],
                 deserializer: Union[DeserializationSchema, KinesisDeserializationSchema],
                 config_props: Dict
                 ):
        gateway = get_gateway()
        j_properties = gateway.jvm.java.util.Properties()
        for key, value in config_props.items():
            j_properties.setProperty(key, value)

        JFlinkKinesisConsumer = gateway.jvm.org.apache.flink.streaming.connectors.kinesis. \
            FlinkKinesisConsumer
        JKinesisDeserializationSchemaWrapper = get_gateway().jvm.org.apache.flink.streaming. \
            connectors.kinesis.serialization.KinesisDeserializationSchemaWrapper

        if isinstance(streams, str):
            streams = [streams]

        if isinstance(deserializer, DeserializationSchema):
            deserializer = JKinesisDeserializationSchemaWrapper(
                deserializer._j_deserialization_schema)

        self._j_kinesis_consumer = JFlinkKinesisConsumer(streams, deserializer, j_properties)
        super(FlinkKinesisConsumer, self).__init__(self._j_kinesis_consumer)

    def set_shard_assigner(self, shard_assigner: KinesisShardAssigner) -> 'FlinkKinesisConsumer':
        """
        Provide a custom assigner to influence how shards are distributed over subtasks.
        """
        self._j_kinesis_consumer.setShardAssigner(shard_assigner._j_kinesis_shard_assigner)
        return self

    def set_periodic_watermark_assigner(
        self,
        periodic_watermark_assigner: AssignerWithPeriodicWatermarksWrapper) \
            -> 'FlinkKinesisConsumer':
        """
        Set the assigner that will extract the timestamp from T and calculate the watermark.
        """
        self._j_kinesis_consumer.setPeriodicWatermarkAssigner(
            periodic_watermark_assigner._j_assigner_with_periodic_watermarks)
        return self

    def set_watermark_tracker(self, watermark_tracker: WatermarkTracker) -> 'FlinkKinesisConsumer':
        """
        Set the global watermark tracker. When set, it will be used by the fetcher to align the
        shard consumers by event time.
        """
        self._j_kinesis_consumer.setWatermarkTracker(watermark_tracker._j_watermark_tracker)
        return self


# ---- KinesisSink ----

class PartitionKeyGenerator(object):
    """
    This is a generator convert from an input element to the partition key, a string.
    """
    def __init__(self, j_partition_key_generator):
        self._j_partition_key_generator = j_partition_key_generator

    @staticmethod
    def fixed() -> 'PartitionKeyGenerator':
        """
        A partitioner ensuring that each internal Flink partition ends up in the same Kinesis
        partition. This is achieved by using the index of the producer task as a PartitionKey.
        """
        return PartitionKeyGenerator(get_gateway().jvm.org.apache.flink.connector.kinesis.table.
                                     FixedKinesisPartitionKeyGenerator())

    @staticmethod
    def random() -> 'PartitionKeyGenerator':
        """
        A PartitionKeyGenerator that maps an arbitrary input element to a random partition ID.
        """
        return PartitionKeyGenerator(get_gateway().jvm.org.apache.flink.connector.kinesis.table.
                                     RandomKinesisPartitionKeyGenerator())


class KinesisStreamsSink(Sink):
    """
    A Kinesis Data Streams (KDS) Sink that performs async requests against a destination stream
    using the buffering protocol.

    The sink internally uses a software.amazon.awssdk.services.kinesis.KinesisAsyncClient to
    communicate with the AWS endpoint.

    The behaviour of the buffering may be specified by providing configuration during the sink
    build time.

    - maxBatchSize: the maximum size of a batch of entries that may be sent to KDS
    - maxInFlightRequests: the maximum number of in flight requests that may exist, if any more in
        flight requests need to be initiated once the maximum has been reached, then it will be
        blocked until some have completed
    - maxBufferedRequests: the maximum number of elements held in the buffer, requests to add
        elements will be blocked while the number of elements in the buffer is at the maximum
    - maxBatchSizeInBytes: the maximum size of a batch of entries that may be sent to KDS
        measured in bytes
    - maxTimeInBufferMS: the maximum amount of time an entry is allowed to live in the buffer,
        if any element reaches this age, the entire buffer will be flushed immediately
    - maxRecordSizeInBytes: the maximum size of a record the sink will accept into the buffer,
        a record of size larger than this will be rejected when passed to the sink
    - failOnError: when an exception is encountered while persisting to Kinesis Data Streams,
        the job will fail immediately if failOnError is set
    """

    def __init__(self, j_kinesis_streams_sink):
        super(KinesisStreamsSink, self).__init__(sink=j_kinesis_streams_sink)

    @staticmethod
    def builder() -> 'KinesisStreamsSinkBuilder':
        return KinesisStreamsSinkBuilder()


class KinesisStreamsSinkBuilder(object):
    """
    Builder to construct KinesisStreamsSink.

    The following example shows the minimum setup to create a KinesisStreamsSink that writes String
    values to a Kinesis Data Streams stream named your_stream_here.

    Example:
    ::

        >>> from pyflink.common.serialization import SimpleStringSchema
        >>> sink_properties = {"aws.region": "eu-west-1"}
        >>> sink = KinesisStreamsSink.builder() \\
        ...     .set_kinesis_client_properties(sink_properties) \\
        ...     .set_stream_name("your_stream_name") \\
        ...     .set_serialization_schema(SimpleStringSchema()) \\
        ...     .set_partition_key_generator(PartitionKeyGenerator.random()) \\
        ...     .build()

    If the following parameters are not set in this builder, the following defaults will be used:

        - maxBatchSize will be 500
        - maxInFlightRequests will be 50
        - maxBufferedRequests will be 10000
        - maxBatchSizeInBytes will be 5 MB i.e. 5 * 1024 * 1024
        - maxTimeInBufferMS will be 5000ms
        - maxRecordSizeInBytes will be 1 MB i.e. 1 * 1024 * 1024
        - failOnError will be false
    """

    def __init__(self):
        JKinesisStreamsSink = get_gateway().jvm.org.apache.flink.connector.kinesis.sink.\
            KinesisStreamsSink
        self._j_kinesis_sink_builder = JKinesisStreamsSink.builder()

    def set_stream_name(self, stream_name: Union[str, List[str]]) -> 'KinesisStreamsSinkBuilder':
        """
        Sets the name of the KDS stream that the sink will connect to. There is no default for this
        parameter, therefore, this must be provided at sink creation time otherwise the build will
        fail.
        """
        self._j_kinesis_sink_builder.setStreamName(stream_name)
        return self

    def set_serialization_schema(self, serialization_schema: SerializationSchema) \
            -> 'KinesisStreamsSinkBuilder':
        """
        Sets the SerializationSchema of the KinesisSinkBuilder.
        """
        self._j_kinesis_sink_builder.setSerializationSchema(
            serialization_schema._j_serialization_schema)
        return self

    def set_partition_key_generator(self, partition_key_generator: PartitionKeyGenerator) \
            -> 'KinesisStreamsSinkBuilder':
        """
        Sets the PartitionKeyGenerator of the KinesisSinkBuilder.
        """
        self._j_kinesis_sink_builder.setPartitionKeyGenerator(
            partition_key_generator._j_partition_key_generator)
        return self

    def set_fail_on_error(self, fail_on_error: bool) -> 'KinesisStreamsSinkBuilder':
        """
        Sets the failOnError of the KinesisSinkBuilder. If failOnError is on, then a runtime
        exception will be raised. Otherwise, those records will be requested in the buffer for
        retry.
        """
        self._j_kinesis_sink_builder.setFailOnError(fail_on_error)
        return self

    def set_kinesis_client_properties(self, kinesis_client_properties: Dict) \
            -> 'KinesisStreamsSinkBuilder':
        """
        Sets the kinesisClientProperties of the KinesisSinkBuilder.
        """
        j_properties = get_gateway().jvm.java.util.Properties()
        for key, value in kinesis_client_properties.items():
            j_properties.setProperty(key, value)
        self._j_kinesis_sink_builder.setKinesisClientProperties(j_properties)
        return self

    def set_max_batch_size(self, max_batch_size: int) -> 'KinesisStreamsSinkBuilder':
        """
        Maximum number of elements that may be passed in a list to be written downstream.
        """
        self._j_kinesis_sink_builder.setMaxBatchSize(max_batch_size)
        return self

    def set_max_in_flight_requests(self, max_in_flight_requests: int) \
            -> 'KinesisStreamsSinkBuilder':
        """
        Maximum number of uncompleted calls to submitRequestEntries that the SinkWriter will allow
        at any given point. Once this point has reached, writes and callbacks to add elements to
        the buffer may block until one or more requests to submitRequestEntries completes.
        """
        self._j_kinesis_sink_builder.setMaxInFlightRequests(max_in_flight_requests)
        return self

    def set_max_buffered_requests(self, max_buffered_requests: int) -> 'KinesisStreamsSinkBuilder':
        """
        The maximum buffer length. Callbacks to add elements to the buffer and calls to write will
        block if this length has been reached and will only unblock if elements from the buffer have
        been removed for flushing.
        """
        self._j_kinesis_sink_builder.setMaxBufferedRequests(max_buffered_requests)
        return self

    def set_max_batch_size_in_bytes(self, max_batch_size_in_bytes: int) \
            -> 'KinesisStreamsSinkBuilder':
        """
        The flush will be attempted if the most recent call to write introduces an element to the
        buffer such that the total size of the buffer is greater than or equal to this threshold
        value. If this happens, the maximum number of elements from the head of the buffer will be
        selected, that is smaller than maxBatchSizeInBytes in size will be flushed.
        """
        self._j_kinesis_sink_builder.setMaxBatchSizeInBytes(max_batch_size_in_bytes)
        return self

    def set_max_time_in_buffer_ms(self, max_time_in_buffer_ms: int) -> 'KinesisStreamsSinkBuilder':
        """
        The maximum amount of time an element may remain in the buffer. In most cases elements are
        flushed as a result of the batch size (in bytes or number) being reached or during a
        snapshot. However, there are scenarios where an element may remain in the buffer forever or
        a long period of time. To mitigate this, a timer is constantly active in the buffer such
        that: while the buffer is not empty, it will flush every maxTimeInBufferMS milliseconds.
        """
        self._j_kinesis_sink_builder.setMaxTimeInBufferMS(max_time_in_buffer_ms)
        return self

    def set_max_record_size_in_bytes(self, max_record_size_in_bytes: int) \
            -> 'KinesisStreamsSinkBuilder':
        """
        The maximum size of each records in bytes. If a record larger than this is passed to the
        sink, it will throw an IllegalArgumentException.
        """
        self._j_kinesis_sink_builder.setMaxRecordSizeInBytes(max_record_size_in_bytes)
        return self

    def build(self) -> 'KinesisStreamsSink':
        """
        Build thd KinesisStreamsSink.
        """
        return KinesisStreamsSink(self._j_kinesis_sink_builder.build())


class KinesisFirehoseSink(Sink):
    """
    A Kinesis Data Firehose (KDF) Sink that performs async requests against a destination delivery
    stream using the buffering protocol.
    """
    def __init__(self, j_kinesis_firehose_sink):
        super(KinesisFirehoseSink, self).__init__(sink=j_kinesis_firehose_sink)

    @staticmethod
    def builder() -> 'KinesisFirehoseSinkBuilder':
        return KinesisFirehoseSinkBuilder()


class KinesisFirehoseSinkBuilder(object):
    """
    Builder to construct KinesisFirehoseSink.

    The following example shows the minimum setup to create a KinesisFirehoseSink that writes
    String values to a Kinesis Data Firehose delivery stream named delivery-stream-name.

    Example:
    ::

        >>> from pyflink.common.serialization import SimpleStringSchema
        >>> sink_properties = {"aws.region": "eu-west-1"}
        >>> sink = KinesisFirehoseSink.builder() \\
        ...     .set_firehose_client_properties(sink_properties) \\
        ...     .set_delivery_stream_name("delivery-stream-name") \\
        ...     .set_serialization_schema(SimpleStringSchema()) \\
        ...     .set_max_batch_size(20) \\
        ...     .build()

    If the following parameters are not set in this builder, the following defaults will be used:
    - maxBatchSize will be 500
    - maxInFlightRequests will be 50
    - maxBufferedRequests will be 10000
    - maxBatchSizeInBytes will be 4 MB i.e. 4 * 1024 * 1024
    - maxTimeInBufferMS will be 5000ms
    - maxRecordSizeInBytes will be 1000 KB i.e. 1000 * 1024
    - failOnError will be false
    """
    def __init__(self):
        JKinesisFirehoseSink = get_gateway().jvm.org.apache.flink.connector.firehose.sink. \
            KinesisFirehoseSink
        self._j_kinesis_sink_builder = JKinesisFirehoseSink.builder()

    def set_delivery_stream_name(self, delivery_stream_name: str) -> 'KinesisFirehoseSinkBuilder':
        """
        Sets the name of the KDF delivery stream that the sink will connect to. There is no default
        for this parameter, therefore, this must be provided at sink creation time otherwise the
        build will fail.
        """
        self._j_kinesis_sink_builder.setDeliveryStreamName(delivery_stream_name)
        return self

    def set_serialization_schema(self, serialization_schema: SerializationSchema) \
            -> 'KinesisFirehoseSinkBuilder':
        """
        Allows the user to specify a serialization schema to serialize each record to persist to
        Firehose.
        """
        self._j_kinesis_sink_builder.setSerializationSchema(
            serialization_schema._j_serialization_schema)
        return self

    def set_fail_on_error(self, fail_on_error: bool) -> 'KinesisFirehoseSinkBuilder':
        """
        If writing to Kinesis Data Firehose results in a partial or full failure being returned,
        the job will fail
        """
        self._j_kinesis_sink_builder.setFailOnError(fail_on_error)
        return self

    def set_firehose_client_properties(self, firehose_client_properties: Dict) \
            -> 'KinesisFirehoseSinkBuilder':
        """
        A set of properties used by the sink to create the firehose client. This may be used to set
        the aws region, credentials etc. See the docs for usage and syntax.
        """
        j_properties = get_gateway().jvm.java.util.Properties()
        for key, value in firehose_client_properties.items():
            j_properties.setProperty(key, value)
        self._j_kinesis_sink_builder.setFirehoseClientProperties(j_properties)
        return self

    def set_max_batch_size(self, max_batch_size: int) -> 'KinesisFirehoseSinkBuilder':
        """
        Maximum number of elements that may be passed in a list to be written downstream.
        """
        self._j_kinesis_sink_builder.setMaxBatchSize(max_batch_size)
        return self

    def set_max_in_flight_requests(self, max_in_flight_requests: int) \
            -> 'KinesisFirehoseSinkBuilder':
        """
        Maximum number of uncompleted calls to submitRequestEntries that the SinkWriter will allow
        at any given point. Once this point has reached, writes and callbacks to add elements to
        the buffer may block until one or more requests to submitRequestEntries completes.
        """
        self._j_kinesis_sink_builder.setMaxInFlightRequests(max_in_flight_requests)
        return self

    def set_max_buffered_requests(self, max_buffered_requests: int) -> 'KinesisFirehoseSinkBuilder':
        """
        The maximum buffer length. Callbacks to add elements to the buffer and calls to write will
        block if this length has been reached and will only unblock if elements from the buffer have
        been removed for flushing.
        """
        self._j_kinesis_sink_builder.setMaxBufferedRequests(max_buffered_requests)
        return self

    def set_max_batch_size_in_bytes(self, max_batch_size_in_bytes: int) \
            -> 'KinesisFirehoseSinkBuilder':
        """
        The flush will be attempted if the most recent call to write introduces an element to the
        buffer such that the total size of the buffer is greater than or equal to this threshold
        value. If this happens, the maximum number of elements from the head of the buffer will be
        selected, that is smaller than maxBatchSizeInBytes in size will be flushed.
        """
        self._j_kinesis_sink_builder.setMaxBatchSizeInBytes(max_batch_size_in_bytes)
        return self

    def set_max_time_in_buffer_ms(self, max_time_in_buffer_ms: int) -> 'KinesisFirehoseSinkBuilder':
        """
        The maximum amount of time an element may remain in the buffer. In most cases elements are
        flushed as a result of the batch size (in bytes or number) being reached or during a
        snapshot. However, there are scenarios where an element may remain in the buffer forever or
        a long period of time. To mitigate this, a timer is constantly active in the buffer such
        that: while the buffer is not empty, it will flush every maxTimeInBufferMS milliseconds.
        """
        self._j_kinesis_sink_builder.setMaxTimeInBufferMS(max_time_in_buffer_ms)
        return self

    def set_max_record_size_in_bytes(self, max_record_size_in_bytes: int) \
            -> 'KinesisFirehoseSinkBuilder':
        """
        The maximum size of each records in bytes. If a record larger than this is passed to the
        sink, it will throw an IllegalArgumentException.
        """
        self._j_kinesis_sink_builder.setMaxRecordSizeInBytes(max_record_size_in_bytes)
        return self

    def build(self) -> 'KinesisFirehoseSink':
        """
        Build thd KinesisFirehoseSink.
        """
        return KinesisFirehoseSink(self._j_kinesis_sink_builder.build())



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from typing import Callable

from pemja import findClass

from pyflink.fn_execution.metrics.embedded.counter_impl import CounterImpl
from pyflink.fn_execution.metrics.embedded.distribution_impl import DistributionImpl
from pyflink.fn_execution.metrics.embedded.meter_impl import MeterImpl
from pyflink.metrics import MetricGroup, Counter, Distribution, Meter

JMeterView = findClass('org.apache.flink.metrics.MeterView')
JMetricGauge = findClass('org.apache.flink.python.metric.embedded.MetricGauge')
JMetricDistribution = findClass('org.apache.flink.python.metric.embedded.MetricDistribution')


class MetricGroupImpl(MetricGroup):

    def __init__(self, metrics):
        self._metrics = metrics

    def add_group(self, name: str, extra: str = None) -> 'MetricGroup':
        if extra is None:
            return MetricGroupImpl(self._metrics.addGroup(name))
        else:
            return MetricGroupImpl(self._metrics.addGroup(name, extra))

    def counter(self, name: str) -> 'Counter':
        return CounterImpl(self._metrics.counter(name))

    def gauge(self, name: str, obj: Callable[[], int]) -> None:
        self._metrics.gauge(name, JMetricGauge(PythonGaugeCallable(obj)))

    def meter(self, name: str, time_span_in_seconds: int = 60) -> 'Meter':
        return MeterImpl(self._metrics.meter(name, JMeterView(time_span_in_seconds)))

    def distribution(self, name: str) -> 'Distribution':
        return DistributionImpl(self._metrics.gauge(name, JMetricDistribution()))


class PythonGaugeCallable(object):
    def __init__(self, func: Callable):
        self.func = func

    def get_value(self):
        return self.func()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import logging
import sys

from pyflink.common import SimpleStringSchema, WatermarkStrategy
from pyflink.datastream import StreamExecutionEnvironment
from pyflink.datastream.connectors.pulsar import PulsarSource, PulsarSink, StartCursor, \
    StopCursor, DeliveryGuarantee, TopicRoutingMode

if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format="%(message)s")

    PULSAR_SQL_CONNECTOR_PATH = 'file:///path/to/flink-sql-connector-pulsar-1.16.0.jar'
    SERVICE_URL = 'pulsar://localhost:6650'
    ADMIN_URL = 'http://localhost:8080'

    env = StreamExecutionEnvironment.get_execution_environment()
    env.set_parallelism(1)
    env.add_jars(PULSAR_SQL_CONNECTOR_PATH)

    pulsar_source = PulsarSource.builder() \
        .set_service_url(SERVICE_URL) \
        .set_admin_url(ADMIN_URL) \
        .set_topics('ada') \
        .set_start_cursor(StartCursor.latest()) \
        .set_unbounded_stop_cursor(StopCursor.never()) \
        .set_subscription_name('pyflink_subscription') \
        .set_deserialization_schema(SimpleStringSchema()) \
        .set_config('pulsar.source.enableAutoAcknowledgeMessage', True) \
        .set_properties({'pulsar.source.autoCommitCursorInterval': '1000'}) \
        .build()

    ds = env.from_source(source=pulsar_source,
                         watermark_strategy=WatermarkStrategy.for_monotonous_timestamps(),
                         source_name="pulsar source")

    pulsar_sink = PulsarSink.builder() \
        .set_service_url(SERVICE_URL) \
        .set_admin_url(ADMIN_URL) \
        .set_producer_name('pyflink_producer') \
        .set_topics('beta') \
        .set_serialization_schema(SimpleStringSchema()) \
        .set_delivery_guarantee(DeliveryGuarantee.AT_LEAST_ONCE) \
        .set_topic_routing_mode(TopicRoutingMode.ROUND_ROBIN) \
        .set_config('pulsar.producer.maxPendingMessages', 1000) \
        .set_properties({'pulsar.producer.batchingMaxMessages': '100'}) \
        .build()

    ds.sink_to(pulsar_sink).name('pulsar sink')

    env.execute()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import logging
import sys

from pyflink.table import (EnvironmentSettings, TableEnvironment, DataTypes)
from pyflink.table.udf import udf


def multi_sink():
    t_env = TableEnvironment.create(EnvironmentSettings.in_streaming_mode())

    table = t_env.from_elements(
        elements=[(1, 'Hello'), (2, 'World'), (3, "Flink"), (4, "PyFlink")],
        schema=['id', 'data'])

    # define the sink tables
    t_env.execute_sql("""
        CREATE TABLE first_sink (
            id BIGINT,
            data VARCHAR
        ) WITH (
            'connector' = 'print'
        )
    """)

    t_env.execute_sql("""
        CREATE TABLE second_sink (
            id BIGINT,
            data VARCHAR
        ) WITH (
            'connector' = 'print'
        )
    """)

    # create a statement set
    statement_set = t_env.create_statement_set()

    # emit the data with id <= 3 to the "first_sink" via sql statement
    statement_set.add_insert_sql("INSERT INTO first_sink SELECT * FROM %s WHERE id <= 3" % table)

    # emit the data which contains "Flink" to the "second_sink"
    @udf(result_type=DataTypes.BOOLEAN())
    def contains_flink(data):
        return "Flink" in data

    second_table = table.where(contains_flink(table.data))
    statement_set.add_insert("second_sink", second_table)

    # execute the statement set
    # remove .wait if submitting to a remote cluster, refer to
    # https://nightlies.apache.org/flink/flink-docs-stable/docs/dev/python/faq/#wait-for-jobs-to-finish-when-executing-jobs-in-mini-cluster
    # for more details
    statement_set.execute().wait()


if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format="%(message)s")

    multi_sink()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from abc import ABC, abstractmethod
from typing import Generic, TypeVar, List, Iterable

from pyflink.common.constants import MAX_LONG_VALUE
from pyflink.datastream.state import StateDescriptor, State, ValueStateDescriptor, \
    ListStateDescriptor, MapStateDescriptor
from pyflink.datastream.window import TimeWindow, CountWindow
from pyflink.fn_execution.datastream.process.timerservice_impl import LegacyInternalTimerServiceImpl
from pyflink.fn_execution.coders import from_type_info, MapCoder, GenericArrayCoder
from pyflink.fn_execution.internal_state import InternalMergingState

K = TypeVar('K')
W = TypeVar('W', TimeWindow, CountWindow)


class Context(Generic[K, W], ABC):
    """
    Information available in an invocation of methods of InternalWindowProcessFunction.
    """

    @abstractmethod
    def get_partitioned_state(self, state_descriptor: StateDescriptor) -> State:
        """
        Creates a partitioned state handle, using the state backend configured for this task.
        """
        pass

    @abstractmethod
    def current_key(self) -> K:
        """
        Returns current key of current processed element.
        """
        pass

    @abstractmethod
    def current_processing_time(self) -> int:
        """
        Returns the current processing time.
        """
        pass

    @abstractmethod
    def current_watermark(self) -> int:
        """
        Returns the current event-time watermark.
        """
        pass

    @abstractmethod
    def get_window_accumulators(self, window: W) -> List:
        """
        Gets the accumulators of the given window.
        """
        pass

    @abstractmethod
    def set_window_accumulators(self, window: W, acc: List):
        """
        Sets the accumulators of the given window.
        """
        pass

    @abstractmethod
    def clear_window_state(self, window: W):
        """
        Clear window state of the given window.
        """
        pass

    @abstractmethod
    def clear_trigger(self, window: W):
        """
        Call Trigger#clear(Window) on trigger.
        """
        pass

    @abstractmethod
    def on_merge(self, new_window: W, merged_windows: Iterable[W]):
        """
        Call Trigger.on_merge() on trigger.
        """
        pass

    @abstractmethod
    def delete_cleanup_timer(self, window: W):
        """
        Deletes the cleanup timer set for the contents of the provided window.
        """
        pass


class WindowContext(Context[K, W]):
    """
    Context of window.
    """

    def __init__(self,
                 window_operator,
                 trigger_context: 'TriggerContext',
                 state_backend,
                 state_value_coder,
                 timer_service: LegacyInternalTimerServiceImpl,
                 is_event_time: bool):
        self._window_operator = window_operator
        self._trigger_context = trigger_context
        self._state_backend = state_backend
        self.timer_service = timer_service
        self.is_event_time = is_event_time
        self.window_state = self._state_backend.get_value_state("window_state", state_value_coder)

    def get_partitioned_state(self, state_descriptor: StateDescriptor) -> State:
        return self._trigger_context.get_partitioned_state(state_descriptor)

    def current_key(self) -> K:
        return self._state_backend.get_current_key()

    def current_processing_time(self) -> int:
        return self.timer_service.current_processing_time()

    def current_watermark(self) -> int:
        return self.timer_service.current_watermark()

    def get_window_accumulators(self, window: W) -> List:
        self.window_state.set_current_namespace(window)
        return self.window_state.value()

    def set_window_accumulators(self, window: W, acc: List):
        self.window_state.set_current_namespace(window)
        self.window_state.update(acc)

    def clear_window_state(self, window: W):
        self.window_state.set_current_namespace(window)
        self.window_state.clear()

    def clear_trigger(self, window: W):
        self._trigger_context.window = window
        self._trigger_context.clear()

    def on_merge(self, new_window: W, merged_windows: Iterable[W]):
        self._trigger_context.window = new_window
        self._trigger_context.merged_windows = merged_windows
        self._trigger_context.on_merge()

    def delete_cleanup_timer(self, window: W):
        cleanup_time = self._window_operator.cleanup_time(window)
        if cleanup_time == MAX_LONG_VALUE:
            # no need to clean up because we didn't set one
            return
        if self.is_event_time:
            self._trigger_context.delete_event_time_timer(cleanup_time)
        else:
            self._trigger_context.delete_processing_time_timer(cleanup_time)


class TriggerContext(object):
    """
    TriggerContext is a utility for handling Trigger invocations. It can be reused by setting the
    key and window fields. No internal state must be kept in the TriggerContext.
    """

    def __init__(self,
                 trigger,
                 timer_service: LegacyInternalTimerServiceImpl[W],
                 state_backend):
        self._trigger = trigger
        self._timer_service = timer_service
        self._state_backend = state_backend
        self.window = None  # type: W
        self.merged_windows = None  # type: Iterable[W]

    def open(self):
        self._trigger.open(self)

    def on_element(self, row, timestamp: int) -> bool:
        return self._trigger.on_element(row, timestamp, self.window)

    def on_processing_time(self, timestamp: int) -> bool:
        return self._trigger.on_processing_time(timestamp, self.window)

    def on_event_time(self, timestamp: int) -> bool:
        return self._trigger.on_event_time(timestamp, self.window)

    def on_merge(self):
        self._trigger.on_merge(self.window, self)

    def get_current_processing_time(self) -> int:
        return self._timer_service.current_processing_time()

    def get_current_watermark(self) -> int:
        return self._timer_service.current_watermark()

    def register_processing_time_timer(self, time: int):
        self._timer_service.register_processing_time_timer(self.window, time)

    def register_event_time_timer(self, time: int):
        self._timer_service.register_event_time_timer(self.window, time)

    def delete_processing_time_timer(self, time: int):
        self._timer_service.delete_processing_time_timer(self.window, time)

    def delete_event_time_timer(self, time: int):
        self._timer_service.delete_event_time_timer(self.window, time)

    def clear(self):
        self._trigger.clear(self.window)

    def get_partitioned_state(self, state_descriptor: StateDescriptor) -> State:
        if isinstance(state_descriptor, ValueStateDescriptor):
            state = self._state_backend.get_value_state(
                state_descriptor.name, from_type_info(state_descriptor.type_info))
        elif isinstance(state_descriptor, ListStateDescriptor):
            array_coder = from_type_info(state_descriptor.type_info)  # type: GenericArrayCoder
            state = self._state_backend.get_list_state(
                state_descriptor.name, array_coder._elem_coder)
        elif isinstance(state_descriptor, MapStateDescriptor):
            map_coder = from_type_info(state_descriptor.type_info)  # type: MapCoder
            key_coder = map_coder._key_coder
            value_coder = map_coder._value_coder
            state = self._state_backend.get_map_state(
                state_descriptor.name, key_coder, value_coder)
        else:
            raise Exception("Unknown supported StateDescriptor %s" % state_descriptor)
        state.set_current_namespace(self.window)
        return state

    def merge_partitioned_state(self, state_descriptor: StateDescriptor):
        if not self.merged_windows:
            state = self.get_partitioned_state(state_descriptor)
            if isinstance(state, InternalMergingState):
                state.merge_namespaces(self.window, self.merged_windows)
            else:
                raise Exception("The given state descriptor does not refer to a mergeable state"
                                " (MergingState)")



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from threading import RLock

import cloudpickle

_lock = RLock()


def loads(payload):
    with _lock:
        # there is race condition for pickle.load() used in multi-thread environment,
        # see https://bugs.python.org/issue36773 for more details.
        return cloudpickle.loads(payload)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from enum import Enum

from pyflink.common import Duration
from pyflink.java_gateway import get_gateway


__all__ = [
    'CassandraSink',
    'ConsistencyLevel',
    'MapperOptions',
    'ClusterBuilder',
    'CassandraCommitter',
    'CassandraFailureHandler'
]

# ---- Classes introduced to construct the MapperOptions ----


class ConsistencyLevel(Enum):
    """
    The consistency level
    """
    ANY = 0
    ONE = 1
    TWO = 2
    THREE = 3
    QUORUM = 4
    ALL = 5
    LOCAL_QUORUM = 6
    EACH_QUORUM = 7
    SERIAL = 8
    LOCAL_SERIAL = 9
    LOCAL_ONE = 10

    def _to_j_consistency_level(self):
        JConsistencyLevel = get_gateway().jvm.com.datastax.driver.core.ConsistencyLevel
        return getattr(JConsistencyLevel, self.name)


class MapperOptions(object):
    """
    This class is used to configure a Mapper after deployment.
    """

    def __init__(self):
        """
        A simple method to construct MapperOptions.

        Example:
        ::

        >>> mapper_option = MapperOptions() \\
        ...    .ttl(1800) \\
        ...    .timestamp(3600) \\
        ...    .consistency_level(ConsistencyLevel.ANY) \\
        ...    .tracing(True) \\
        ...    .save_null_fields(True)
        """
        JSimpleMapperOptions = get_gateway().jvm.org.apache.flink.streaming.connectors. \
            cassandra.SimpleMapperOptions
        self._j_mapper_options = JSimpleMapperOptions()

    def ttl(self, ttl: int) -> 'MapperOptions':
        """
        Creates a new Option object to add time-to-live to a mapper operation. This is only
        valid for save operations.
        """
        self._j_mapper_options.ttl(ttl)
        return self

    def timestamp(self, timestamp: int) -> 'MapperOptions':
        """
        Creates a new Option object to add a timestamp to a mapper operation. This is only
        valid for save and delete operations.
        """
        self._j_mapper_options.timestamp(timestamp)
        return self

    def consistency_level(self, cl: ConsistencyLevel) -> 'MapperOptions':
        """
        Creates a new Option object to add a consistency level value to a mapper operation.
        This is valid for save, delete and get operations.
        """
        self._j_mapper_options.consistencyLevel(cl._to_j_consistency_level())
        return self

    def tracing(self, enabled: bool) -> 'MapperOptions':
        """
        Creates a new Option object to enable query tracing for a mapper operation. This is
        valid for save, delete and get operations.
        """
        self._j_mapper_options.tracing(enabled)
        return self

    def save_null_fields(self, enabled: bool) -> 'MapperOptions':
        """
        Creates a new Option object to specify whether null entity fields should be included in
        insert queries. This option is valid only for save operations.
        """
        self._j_mapper_options.saveNullFields(enabled)
        return self

    def if_not_exists(self, enabled: bool) -> 'MapperOptions':
        """
        Creates a new Option object to specify whether an IF NOT EXISTS clause should be included in
        insert queries. This option is valid only for save operations.

        If this option is not specified, it defaults to false (IF NOT EXISTS statements are not
        used).
        """
        self._j_mapper_options.ifNotExists(enabled)
        return self


class ClusterBuilder(object):
    """
    This class is used to configure a Cluster after deployment. The cluster represents the
    connection that will be established to Cassandra.
    """

    def __init__(self, j_cluster_builder):
        self._j_cluster_builder = j_cluster_builder


class CassandraCommitter(object):
    """
    CheckpointCommitter that saves information about completed checkpoints within a separate table
    in a cassandra database.
    """

    def __init__(self, j_checkpoint_committer):
        self._j_checkpoint_committer = j_checkpoint_committer

    @staticmethod
    def default_checkpoint_committer(builder: ClusterBuilder, key_space: str = None) \
            -> 'CassandraCommitter':
        """
        CheckpointCommitter that saves information about completed checkpoints within a separate
        table in a cassandra database.

        Entries are in the form: | operator_id | subtask_id | last_completed_checkpoint |
        """
        JCassandraCommitter = get_gateway().jvm.org.apache.flink.streaming.connectors.\
            cassandra.CassandraCommitter
        if key_space is None:
            j_checkpoint_committer = JCassandraCommitter(builder._j_cluster_builder)
        else:
            j_checkpoint_committer = JCassandraCommitter(builder._j_cluster_builder, key_space)
        return CassandraCommitter(j_checkpoint_committer)


class CassandraFailureHandler(object):
    """
    Handle a failed Throwable.
    """

    def __init__(self, j_cassandra_failure_handler):
        self._j_cassandra_failure_handler = j_cassandra_failure_handler

    @staticmethod
    def no_op() -> 'CassandraFailureHandler':
        """
        A CassandraFailureHandler that simply fails the sink on any failures.
        This is also the default failure handler if not specified.
        """
        return CassandraFailureHandler(get_gateway().jvm.org.apache.flink.streaming.connectors.
                                       cassandra.NoOpCassandraFailureHandler())


# ---- CassandraSink ----


class CassandraSink(object):
    """
    Sets the ClusterBuilder for this sink. A ClusterBuilder is used to configure the connection to
    cassandra.
    """

    def __init__(self, j_cassandra_sink):
        self._j_cassandra_sink = j_cassandra_sink

    def name(self, name: str) -> 'CassandraSink':
        """
        Set the name of this sink. This name is used by the visualization and logging during
        runtime.
        """
        self._j_cassandra_sink.name(name)
        return self

    def uid(self, uid: str) -> 'CassandraSink':
        """
        Sets an ID for this operator. The specified ID is used to assign the same operator ID
        across job submissions (for example when starting a job from a savepoint).
        Note that this ID needs to be unique per transformation and job. Otherwise, job submission
        will fail.
        """
        self._j_cassandra_sink.uid(uid)
        return self

    def set_uid_hash(self, uid_hash: str) -> 'CassandraSink':
        """
        Sets an user provided hash for this operator. This will be used AS IS the create the
        JobVertexID.

        The user provided hash is an alternative to the generated hashes, that is considered when
        identifying an operator through the default hash mechanics fails (e.g. because of changes
        between Flink versions).

        Note that this should be used as a workaround or for trouble shooting. The provided hash
        needs to be unique per transformation and job. Otherwise, job submission will fail.
        Furthermore, you cannot assign user-specified hash to intermediate nodes in an operator
        chain and trying so will let your job fail.

        A use case for this is in migration between Flink versions or changing the jobs in a way
        that changes the automatically generated hashes. In this case, providing the previous hashes
        directly through this method (e.g. obtained from old logs) can help to reestablish a lost
        mapping from states to their target operator.
        """
        self._j_cassandra_sink.setUidHash(uid_hash)
        return self

    def set_parallelism(self, parallelism: int) -> 'CassandraSink':
        """
        Sets the parallelism for this sink. The degree must be higher than zero.
        """
        self._j_cassandra_sink.setParallelism(parallelism)
        return self

    def disable_chaining(self) -> 'CassandraSink':
        """
        Turns off chaining for this operator so thread co-location will not be used as an
        optimization.
        """
        self._j_cassandra_sink.disableChaining()
        return self

    def slot_sharing_group(self, slot_sharing_group: str) -> 'CassandraSink':
        """
        Sets the slot sharing group of this operation. Parallel instances of operations that are in
        the same slot sharing group will be co-located in the same TaskManager slot, if possible.

        Operations inherit the slot sharing group of input operations if all input operations are in
        the same slot sharing group and no slot sharing group was explicitly specified.

        Initially an operation is in the default slot sharing group. An operation can be put into
        the default group explicitly by setting the slot sharing group to {@code "default"}.
        """
        self._j_cassandra_sink.slotSharingGroup(slot_sharing_group)
        return self

    @staticmethod
    def add_sink(input) -> 'CassandraSinkBuilder':
        """
        Writes a DataStream into a Cassandra database.
        """
        JCassandraSink = get_gateway().jvm \
            .org.apache.flink.streaming.connectors.cassandra.CassandraSink
        j_cassandra_sink_builder = JCassandraSink.addSink(input._j_data_stream)
        return CassandraSink.CassandraSinkBuilder(j_cassandra_sink_builder)

    class CassandraSinkBuilder(object):
        """
        Builder for a CassandraSink.
        """

        def __init__(self, j_cassandra_sink_builder):
            self._j_cassandra_sink_builder = j_cassandra_sink_builder

        def set_query(self, query: str) -> 'CassandraSink.CassandraSinkBuilder':
            """
            Sets the query that is to be executed for every record.
            """
            self._j_cassandra_sink_builder.setQuery(query)
            return self

        def set_host(self, host: str, port: int = 9042) -> 'CassandraSink.CassandraSinkBuilder':
            """
            Sets the cassandra host/port to connect to.
            """
            self._j_cassandra_sink_builder.setHost(host, port)
            return self

        def set_cluster_builder(self, builder: ClusterBuilder) \
                -> 'CassandraSink.CassandraSinkBuilder':
            """
            Sets the ClusterBuilder for this sink. A ClusterBuilder is used to configure the
            connection to cassandra.
            """
            self._j_cassandra_sink_builder.setClusterBuilder(builder._j_cluster_builder)
            return self

        def enable_write_ahead_log(self, committer: CassandraCommitter = None) \
                -> 'CassandraSink.CassandraSinkBuilder':
            """
            Enables the write-ahead log, which allows exactly-once processing for non-deterministic
            algorithms that use idempotent updates.
            """
            if committer is None:
                self._j_cassandra_sink_builder.enableWriteAheadLog()
            else:
                self._j_cassandra_sink_builder.enableWriteAheadLog(
                    committer._j_checkpoint_committer)
            return self

        def set_mapper_options(self, options: MapperOptions) \
                -> 'CassandraSink.CassandraSinkBuilder':
            """
            Sets the mapper options for this sink. The mapper options are used to configure the
            DataStax com.datastax.driver.mapping.Mapper when writing POJOs.
            This call has no effect if the input DataStream for this sink does not contain POJOs.
            """
            self._j_cassandra_sink_builder.setMapperOptions(options._j_mapper_options)
            return self

        def set_failure_handler(self, failure_handler: CassandraFailureHandler) \
                -> 'CassandraSink.CassandraSinkBuilder':
            """
            Sets the failure handler for this sink. The failure handler is used to provide custom
            error handling.
            """
            self._j_cassandra_sink_builder.setFailureHandler(
                failure_handler._j_cassandra_failure_handler)
            return self

        def set_max_concurrent_requests(self,
                                        max_concurrent_requests: int,
                                        duration: Duration = None) \
                -> 'CassandraSink.CassandraSinkBuilder':
            """
            Sets the maximum allowed number of concurrent requests for this sink.
            """
            if duration is None:
                self._j_cassandra_sink_builder.setMaxConcurrentRequests(max_concurrent_requests)
            else:
                self._j_cassandra_sink_builder.setMaxConcurrentRequests(
                    max_concurrent_requests, duration._j_duration)
            return self

        def enable_ignore_null_fields(self) -> 'CassandraSink.CassandraSinkBuilder':
            """
            Enables ignoring null values, treats null values as unset and avoids writing null fields
            and creating tombstones.
            This call has no effect if CassandraSinkBuilder.enableWriteAheadLog() is called.
            """
            self._j_cassandra_sink_builder.enableIgnoreNullFields()
            return self

        def build(self) -> 'CassandraSink':
            """
            Finalizes the configuration of this sink.
            """
            return CassandraSink(self._j_cassandra_sink_builder.build())
